@@ -18,6 +18,12 @@
 """An adapter between a Git control dir and a Bazaar ControlDir."""
 
 import contextlib
+import os
+
+try:
+    from dulwich.refs import SymrefLoop
+except ImportError:
+    SymrefLoop = KeyError
 
 from .. import (
     branch as _mod_branch,
@@ -29,6 +35,8 @@ from .. import (
 from ..transport import (
     do_catching_redirections,
     get_transport_from_path,
+    FileExists,
+    NoSuchFile,
     )
 
 from ..controldir import (
@@ -207,14 +215,19 @@ class GitDir(ControlDir):
                     subtrees = []
                 for path in subtrees:
                     target = urlutils.join(url, urlutils.escape(path))
-                    sublocation = wt.reference_parent(
-                        path, possible_transports=possible_transports)
+                    sublocation = wt.get_reference_info(path)
                     if sublocation is None:
-                        trace.warning(
-                            'Ignoring nested tree %s, parent location unknown.',
-                            path)
+                        trace.warning("Unable to find submodule info for %s", path)
                         continue
-                    sublocation.controldir.sprout(
+                    remote_url = urlutils.join(self.user_url, sublocation)
+                    try:
+                        subbranch = _mod_branch.Branch.open(remote_url, possible_transports=possible_transports)
+                    except brz_errors.NotBranchError as e:
+                        trace.warning(
+                            'Unable to clone submodule %s from %s: %s',
+                            path, remote_url, e)
+                        continue
+                    subbranch.controldir.sprout(
                         target, basis.get_reference_revision(path),
                         force_new_repo=force_new_repo, recurse=recurse,
                         stacked=stacked)
@@ -457,10 +470,10 @@ class LocalGitControlDirFormat(GitControlDirFormat):
         try:
             transport = do_catching_redirections(
                 make_directory, transport, redirected)
-        except brz_errors.FileExists:
+        except FileExists:
             if not use_existing_dir:
                 raise
-        except brz_errors.NoSuchFile:
+        except NoSuchFile:
             if not create_prefix:
                 raise
             transport.create_prefix()
@@ -579,7 +592,10 @@ class LocalGitDir(GitDir):
 
     def get_branch_reference(self, name=None):
         ref = self._get_selected_ref(name)
-        target_ref = self._get_symref(ref)
+        try:
+            target_ref = self._get_symref(ref)
+        except SymrefLoop:
+            raise BranchReferenceLoop(self)
         if target_ref is not None:
             from .refs import ref_to_branch_name
             try:
@@ -594,7 +610,7 @@ class LocalGitDir(GitDir):
                     params = {}
             try:
                 commondir = self.control_transport.get_bytes('commondir')
-            except brz_errors.NoSuchFile:
+            except NoSuchFile:
                 base_url = self.user_url.rstrip('/')
             else:
                 base_url = urlutils.local_path_to_url(
@@ -640,10 +656,8 @@ class LocalGitDir(GitDir):
                 self.root_transport.base, controldir=self)
         try:
             ref_chain, unused_sha = self._git.refs.follow(ref)
-        except KeyError as e:
-            raise brz_errors.NotBranchError(
-                self.root_transport.base, controldir=self,
-                detail='intermediate ref %s missing' % e.args[0])
+        except SymrefLoop as e:
+            raise BranchReferenceLoop(self)
         if ref_chain[-1] == b'HEAD':
             controldir = self
         else:
@@ -799,9 +813,9 @@ class LocalGitDir(GitDir):
     def _find_commondir(self):
         try:
             commondir = self.control_transport.get_bytes('commondir')
-        except brz_errors.NoSuchFile:
+        except NoSuchFile:
             return self
         else:
-            commondir = commondir.rstrip(b'/.git/').decode(osutils._fs_enc)
+            commondir = os.fsdecode(commondir.rstrip(b'/.git/'))
             return ControlDir.open_from_transport(
                 get_transport_from_path(commondir))
